from dataclasses import dataclass

from battlebit_community_server_api.model.LeaningSide import LeaningSide
from battlebit_community_server_api.model.LoadoutIndex import LoadoutIndex
from battlebit_community_server_api.model.PlayerStand import PlayerStand
from battlebit_community_server_api.model.SteamId import SteamId
from battlebit_community_server_api.model.Vector3 import Vector3


@dataclass
class PlayerState:
    steam_id: SteamId
    position: Vector3
    standing: PlayerStand
    leaning_side: LeaningSide
    loadout_index: LoadoutIndex
    is_in_seat: bool
    is_bleeding: bool
    ping: int



import asyncio
import logging
import time
from threading import Thread

from battlebit_community_server_api.ApiServer import ApiServer, build_api_server
from battlebit_community_server_api.model.OpCodes import OpCodes
from battlebit_community_server_api.model.OutgoingGameServerMessage import OutgoingGameServerMessage
from battlebit_community_server_api.model.RoomSettings import RoomSettings

event_loop = asyncio.new_event_loop()
my_api_server: ApiServer = build_api_server("0.0.0.0", 30000, event_loop)

# Construct new room settings
new_room_settings = RoomSettings(
    damage_multiplier=1.0,
    friendly_fire_enabled=False,
    hide_map_votes=True,
    only_winner_team_can_vote=True,
    player_collision_enabled=True,

    medic_limit_per_squad=5,
    engineer_limit_per_squad=6,
    support_limit_per_squad=7,
    recon_limit_per_squad=2,

    can_vote_for_day=True,
    can_vote_for_night=True,

    tank_spawn_delay_multiplier=1.0,
    transport_spawn_delay_multiplier=1.0,
    sea_vehicle_spawn_delay_multiplier=1.0,
    apc_spawn_delay_multiplier=1.0,
    helicopter_spawn_delay_multiplier=1.0,

    unlock_all_attachments=True,
    teamless_mode_enabled=False,
    squad_required_to_change_role=False
)

# Prepare an outgoing server message to be added to the queue
my_outgoing_game_server_message = OutgoingGameServerMessage(
    op_code=OpCodes.SET_NEW_ROOM_SETTINGS,
    value=new_room_settings.to_bytes()
)


logging.basicConfig(level=logging.DEBUG)  # Optional: If you want to have an in-depth view into what's going on.

# Adding OutgoingGameServerMessages to the queue from outside the main loop (triggered by surrounding code)
# required the API to be run in a Thread, so it doesn't block.
api_thread = Thread(target=my_api_server.start, daemon=False)
api_thread.start()
SLEEP_TIME = 3
for i in range(SLEEP_TIME):
    logging.debug(f"Sleeping: {SLEEP_TIME - i}")
    time.sleep(1)

logging.debug(f"Adding new room settings message to queue")

# It is PARAMOUNT that all methods on the API server are called threadsafe. Otherwise, you might
# run into race conditions or locks. Use asyncio.run_coroutine_threadsafe whenever possible !
asyncio.run_coroutine_threadsafe(
    my_api_server.add_outgoing_game_server_message_to_queue(my_outgoing_game_server_message),
    event_loop
)



from dataclasses import dataclass
from typing import Optional

from battlebit_community_server_api.model.PlayerBody import PlayerBody
from battlebit_community_server_api.model.ReasonOfDamage import ReasonOfDamage
from battlebit_community_server_api.model.SteamId import SteamId
from battlebit_community_server_api.model.Vector3 import Vector3


@dataclass
class KillReport:
    killer: SteamId
    killer_pos: Vector3
    victim: SteamId
    victim_pos: Vector3
    tool_name: Optional[str]
    body_part: PlayerBody
    damage_reason: ReasonOfDamage



class IpAddress:
    def __init__(self, uint32: int) -> None:
        self._uint32 = uint32

    def __str__(self) -> str:
        byte1 = (self._uint32 >> 24) & 0xFF
        byte2 = (self._uint32 >> 16) & 0xFF
        byte3 = (self._uint32 >> 8) & 0xFF
        byte4 = self._uint32 & 0xFF
        return f"{byte4}.{byte3}.{byte2}.{byte1}"

    def __int__(self) -> int:
        return self._uint32



import struct
from dataclasses import dataclass

from battlebit_community_server_api.model.PlayerLoadout import PlayerLoadout
from battlebit_community_server_api.model.PlayerSpawningPosition import PlayerSpawningPosition
from battlebit_community_server_api.model.PlayerStand import PlayerStand
from battlebit_community_server_api.model.PlayerWearings import PlayerWearings
from battlebit_community_server_api.model.Vector3 import Vector3


@dataclass
class PlayerSpawnArguments:
    position: PlayerSpawningPosition
    loadout: PlayerLoadout
    wearings: PlayerWearings
    vector_position: Vector3
    look_direction: Vector3
    stand: PlayerStand
    spawn_protection: float

    def to_bytes(self) -> bytes:
        out = self.position.as_bytes()
        out += self.loadout.as_bytes()
        out += self.wearings.as_bytes()
        out += self.vector_position.as_bytes()
        out += self.look_direction.as_bytes()
        out += self.stand.as_bytes()
        out += struct.pack("f", self.spawn_protection)

        return out



import logging

from battlebit_community_server_api.ApiServer import ApiServer, build_api_server
from battlebit_community_server_api.model.OpCodes import OpCodes
from battlebit_community_server_api.model.OutgoingGameServerMessage import OutgoingGameServerMessage
from battlebit_community_server_api.model.PlayerJoiningArguments import PlayerJoiningArguments
from battlebit_community_server_api.model.Squads import Squads
from battlebit_community_server_api.model.Team import Team
from battlebit_community_server_api.service.TcpParsingService import TcpParsingService

my_api_server: ApiServer = build_api_server("0.0.0.0", 30000, None)


async def my_custom_on_player_joining(data: bytes) -> bytes:
    """
    This function:

    1. parses the data received by the "On Player Joining" event.
    2. Sets the players rank to 150.
    3. Builds the appropriate response message.
    4. Returns the response as bytes.
    """
    steam_id, player_stats = TcpParsingService.parse_on_player_joining(data)
    player_stats.progress.rank = 150
    response = OutgoingGameServerMessage(op_code=OpCodes.SEND_PLAYER_STATS, value=steam_id.to_bytes())
    response.add_bytes(PlayerJoiningArguments(player_stats, squad=Squads.NO_SQUAD, team=Team.NONE).to_bytes())
    return response.serialize()


logging.basicConfig(level=logging.DEBUG)  # Optional: If you want to have an in-depth view into what's going on.
my_api_server.register_handler(OpCodes.ON_PLAYER_JOINING, my_custom_on_player_joining)

my_api_server.start()



import logging
from dataclasses import dataclass
from typing import Optional

from battlebit_community_server_api.model.GameRole import GameRole
from battlebit_community_server_api.model.SteamId import SteamId
from battlebit_community_server_api.model.Team import Team
from battlebit_community_server_api.model.Squads import Squads
from battlebit_community_server_api.model.Map import Map
from battlebit_community_server_api.model.Gamemode import Gamemode


@dataclass
class Command:
    command_string: str
    arguments: list[str]

    def __len__(self) -> int:
        return len(self.as_string())

    def as_string(self) -> str:
        return f"{self.command_string} {' '.join(self.arguments).rstrip(' ')}".rstrip(' ')


class ForceStartGame(Command):
    def __init__(self) -> None:
        self.command_string = "forcestart"
        self.arguments = []


class ForceEndGame(Command):
    def __init__(self) -> None:
        self.command_string = "endgame"
        self.arguments = []


class SayToAllChat(Command):
    def __init__(self, message: str) -> None:
        self.command_string = "say"
        self.arguments = message.split(" ")


class SayToChat(Command):
    def __init__(self, message: str, target_steam_id: SteamId) -> None:
        self.command_string = "sayto"
        self.arguments = [str(target_steam_id.steam_id)]
        self.arguments += message.split(" ")


class SetRoleTo(Command):
    def __init__(self, steam_id: SteamId, role: GameRole) -> None:
        self.command_string = "setrole"
        self.arguments = [str(steam_id), str(role.value)]


class SetTeamTo(Command):
    def __init__(self, steam_id: SteamId, team: Team) -> None:
        self.command_string = "changeteam"
        self.arguments = [str(steam_id), "a" if team == Team.US else "b"]


class SetNewPassword(Command):
    def __init__(self, new_password: str) -> None:
        self.command_string = "setpass"
        self.arguments = [new_password]


class SetPingLimit(Command):
    def __init__(self, new_ping_limit: int) -> None:
        self.command_string = "setmaxping"
        self.arguments = [str(new_ping_limit)]


class AnnounceShort(Command):
    def __init__(self, announcement_text: str) -> None:
        self.command_string = "an"
        self.arguments = [announcement_text]


class AnnounceLong(Command):
    def __init__(self, announcement_text: str) -> None:
        self.command_string = "ann"
        self.arguments = [announcement_text]


class UILogOnServer(Command):
    def __init__(self, log_message: str, message_lifetime_seconds: float) -> None:
        self.command_string = "serverlog"
        self.arguments = [log_message, str(message_lifetime_seconds)]


class SetLoadingScreenText(Command):
    def __init__(self, new_loading_screen_text: str) -> None:
        self.command_string = "setloadingscreentext"
        self.arguments = [new_loading_screen_text]


class SetRulesScreenText(Command):
    def __init__(self, new_rules_screen_text: str) -> None:
        self.command_string = "setrulesscreentext"
        self.arguments = [new_rules_screen_text]


class StopServer(Command):
    def __init__(self) -> None:
        self.command_string = "stop"
        self.arguments = []


class CloseServer(Command):
    def __init__(self) -> None:
        self.command_string = "notifyend"
        self.arguments = []


class KickAllPlayers(Command):
    def __init__(self) -> None:
        self.command_string = "kick all"
        self.arguments = []


class KickPlayer(Command):
    def __init__(self, steam_id: SteamId, reason_for_kick: Optional[str] = "") -> None:
        self.command_string = "kick"
        self.arguments = [str(steam_id.steam_id), reason_for_kick]


class KillPlayer(Command):
    def __init__(self, steam_id: SteamId) -> None:
        self.command_string = "kill"
        self.arguments = [str(steam_id.steam_id)]


class SwapTeam(Command):
    def __init__(self, steam_id: SteamId) -> None:
        self.command_string = "changeteam"
        self.arguments = [str(steam_id.steam_id)]


class KickFromSquad(Command):
    def __init__(self, steam_id: SteamId) -> None:
        self.command_string = "squadkick"
        self.arguments = [str(steam_id.steam_id)]


class JoinSquad(Command):
    def __init__(self, steam_id: SteamId, squad: Squads) -> None:
        self.command_string = "setsquad"
        self.arguments = [str(steam_id.steam_id), squad.value]


class DisbandPlayerSquad(Command):
    def __init__(self, steam_id: SteamId) -> None:
        self.command_string = "squaddisband"
        self.arguments = [str(steam_id.steam_id)]


class PromoteSquadLeader(Command):
    def __init__(self, steam_id: SteamId) -> None:
        self.command_string = "squadpromote"
        self.arguments = [str(steam_id.steam_id)]


class WarnPlayer(Command):
    def __init__(self, steam_id: SteamId, warn_message: str) -> None:
        self.command_string = "warn"
        self.arguments = [str(steam_id.steam_id), warn_message]


class MessageToPlayer(Command):
    def __init__(self, steam_id: SteamId, message: str, fade_out_time_seconds: Optional[float] = None) -> None:
        if not fade_out_time_seconds:
            self.command_string = "msg"
            self.arguments = [str(steam_id.steam_id), message]
        else:
            self.command_string = "msgf"
            self.arguments = [str(steam_id.steam_id), str(fade_out_time_seconds), message]


class SetPlayerHp(Command):
    def __init__(self, steam_id: SteamId, hit_points: float) -> None:
        self.command_string = "sethp"
        self.arguments = [str(steam_id.steam_id), str(hit_points)]


class DamagePlayer(Command):
    def __init__(self, steam_id: SteamId, damage_amount: float) -> None:
        self.command_string = "givedamage"
        self.arguments = [str(steam_id.steam_id), str(damage_amount)]


class HealPlayer(Command):
    def __init__(self, steam_id: SteamId, heal_amount: float) -> None:
        self.command_string = "heal"
        self.arguments = [str(steam_id.steam_id), str(heal_amount)]


class SetSquadPoints(Command):
    def __init__(self, team: Team, squad: Squads, points: int) -> None:
        self.command_string = "setsquadpoints"
        self.arguments = [str(team.value), str(squad.value), str(points)]


class SetMapRotation(Command):
    def __init__(self, maps: list[Map]) -> None:
        if len(maps) > 8:
            logging.warning("Can not set more than 8 maps, trimming tail...")
            maps = maps[:8]
        self.command_string = "setmaprotation"
        self.arguments = [",".join(maps)]


class SetGamemodeRotation(Command):
    def __init__(self, gamemodes: list[Gamemode]) -> None:
        if len(gamemodes) > 8:
            logging.warning("Can not set more than 8 gamemodes, trimming tail...")
            gamemodes = gamemodes[:8]
        self.command_string = "setgamemoderotation"
        self.arguments = [",".join(gamemodes)]



import asyncio
import datetime
import logging
import struct
from typing import Callable, Optional, Union

from battlebit_community_server_api.command.Command import Command
from battlebit_community_server_api.handler.DefaultHandlers import DefaultHandlers
from battlebit_community_server_api.helper.StructHelper import read_uint8
from battlebit_community_server_api.model.OutgoingGameServerMessage import OutgoingGameServerMessage
from battlebit_community_server_api.service.TcpParsingService import TcpParsingService
from battlebit_community_server_api.model.OpCodes import OpCodes


class GameServerCommunicationService:
    """
    The GameServerCommunicationService is responsible for establishing and maintaining the TCP socket connection
    to the GameServer. It also forwards raw data messages in both directions.
    """
    PACKET_SIZE = 1024 * 1024 * 4  # 4MB
    KEEP_ALIVE_FREQUENCY = 15  # Seconds
    KEEP_ALIVE_MESSAGE = b"\x00\x00\x00\x00"
    TIMEOUT_AFTER = 60  # Seconds

    def __init__(self, host: str, port: int) -> None:
        self._host = host
        self._port = port
        self._socket_server: Optional[asyncio.Server] = None
        self._handler_register: dict[OpCodes, Callable] = {}
        self._queue = set()
        self._writer: Optional[asyncio.StreamWriter] = None
        self._reader: Optional[asyncio.StreamReader] = None

    async def start(self) -> None:
        self._socket_server = await asyncio.start_server(self._handle_incoming_connection, self._host, self._port,
                                                         limit=self.PACKET_SIZE, start_serving=False)
        await self._socket_server.serve_forever()

    async def _handle_incoming_connection(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter) -> None:
        logging.debug(f"New connection from: {writer.get_extra_info('peername')}")
        self._writer, self._reader = writer, reader
        await self._handle_incoming_data()
        logging.debug(f"Connection with {writer.get_extra_info('peername')} has been closed!")

    async def _handle_incoming_data(self) -> None:
        # ToDo: Handle OSError
        last_packet_sent = datetime.datetime.now()
        last_packet_received = datetime.datetime.now()
        while True:
            # Setting current timestamp
            now = datetime.datetime.now()
            # Wait for new packet
            try:
                data = await self._reader.read(self.PACKET_SIZE)
            except ConnectionResetError:
                logging.error("ConnectionReset!!")
                self._writer, self._reader = None, None
                return
            # Special case: Receiving HAIL
            if read_uint8(data)[0] == OpCodes.HAIL:
                logging.debug("Received HAIL...")
                try:
                    self._writer.write(await self._handler_register[OpCodes.HAIL](data))
                except KeyError:
                    self._writer.write(await DefaultHandlers.get_default_handler_by_op_code(OpCodes.HAIL)(data))
                await self._writer.drain()
                continue
            # Unpack received data in messages
            messages = TcpParsingService.split_data_into_individual_messages(data)
            # Handling possible timeout scenario
            if not messages and last_packet_received + datetime.timedelta(seconds=self.TIMEOUT_AFTER) < now:
                logging.error("Gameserver closed the connection")
                return
            last_packet_received = now  # Updating timestamp
            # Triggering handlers for messages
            for message in messages:
                op_code, bytes_ = message
                # Logging
                if op_code not in (OpCodes.GAME_TICK, OpCodes.NOTIFY_NEW_ROUND_STATE):
                    logging.debug(f"Received {op_code.name} message...")
                # Check for custom handlers
                if op_code in self._handler_register.keys():
                    handler_result = await self._handler_register[op_code](bytes_)
                # Execute default handler if no custom handler present
                else:
                    try:
                        handler_result = await DefaultHandlers.get_default_handler_by_op_code(op_code)(bytes_)
                    except NotImplementedError as err:
                        logging.error(err)
                        continue

                # If handler returned any result
                if handler_result:
                    last_packet_sent = now
                    self._writer.write(handler_result)
                    await self._writer.drain()

            # Checking if keepalive message is needed, and send one if it is.
            if last_packet_sent + datetime.timedelta(seconds=self.KEEP_ALIVE_FREQUENCY) < datetime.datetime.now():
                last_packet_sent = now
                self._writer.write(self.KEEP_ALIVE_MESSAGE)
                await self._writer.drain()

    async def _execute_command_or_outgoing_game_server_message(self, command_or_msg: Union[Command, OutgoingGameServerMessage]) -> None:
        if hasattr(command_or_msg, "command_string"):  # FixMe: isinstance() seems to not be working
            command = command_or_msg
            logging.debug(f"Executing command: {command}")
            if not self._writer:
                # ToDo: Shit way of doing this
                await asyncio.sleep(1)
                await self._execute_command_or_outgoing_game_server_message(command)
            else:
                message = OutgoingGameServerMessage(op_code=OpCodes.EXECUTE_COMMAND)
                message.add_bytes(struct.pack("H", len(command)))
                message.add_string(command.as_string())
                self._writer.write(message.serialize())
                await self._writer.drain()
        elif hasattr(command_or_msg, "serialize"):
            msg = command_or_msg
            logging.debug(f"Executing outgoing game server message: {msg}")
            if not self._writer:
                # ToDo: Shit way of doing this
                await asyncio.sleep(1)
                await self._execute_command_or_outgoing_game_server_message(msg)
            else:
                self._writer.write(msg.serialize())
                await self._writer.drain()
        else:
            logging.warning(f"Unknown type queued for execution: {type(command_or_msg)}")

    def register_handler(self, op_code: OpCodes, handler: Callable) -> None:
        self._handler_register[op_code] = handler

    async def add_command_to_queue(self, command: Command) -> None:
        task = asyncio.create_task(self._execute_command_or_outgoing_game_server_message(command))
        self._queue.add(task)
        task.add_done_callback(self._queue.discard)

    async def add_outgoing_game_server_message_to_queue(self, msg: OutgoingGameServerMessage) -> None:
        task = asyncio.create_task(self._execute_command_or_outgoing_game_server_message(msg))
        self._queue.add(task)
        task.add_done_callback(self._queue.discard)



from dataclasses import dataclass


@dataclass
class SteamId:
    steam_id: int

    def to_bytes(self) -> bytes:
        return int(self.steam_id).to_bytes(length=8, byteorder="little")

    def __str__(self) -> str:
        return f"{self.steam_id}"

    def __repr__(self) -> str:
        return f"{self.steam_id}"



import logging
from typing import Optional

from battlebit_community_server_api.helper.StructHelper import *
from battlebit_community_server_api.model import PlayerProgress
from battlebit_community_server_api.model.ChatChannel import ChatChannel
from battlebit_community_server_api.model.GameRole import GameRole
from battlebit_community_server_api.model.MapSize import MapSize
from battlebit_community_server_api.model.OpCodes import OpCodes
from battlebit_community_server_api.model.GameState import GameState
from battlebit_community_server_api.model.PlayerLoadout import read_from_bytes as build_player_loadout, PlayerLoadout
from battlebit_community_server_api.model.PlayerSpawnArguments import PlayerSpawnArguments
from battlebit_community_server_api.model.PlayerSpawningPosition import PlayerSpawningPosition
from battlebit_community_server_api.model.PlayerStand import PlayerStand
from battlebit_community_server_api.model.PlayerStats import PlayerStats
from battlebit_community_server_api.model.Role import Role
from battlebit_community_server_api.model.RoomSettings import from_bytes as room_settings_from_bytes
from battlebit_community_server_api.model.ServerInfo import ServerInfo
from battlebit_community_server_api.model.Squads import Squads
from battlebit_community_server_api.model.SteamId import SteamId
from battlebit_community_server_api.model.Team import Team
from battlebit_community_server_api.model.PlayerWearings import read_from_bytes as build_player_wearings, PlayerWearings
from battlebit_community_server_api.model.Vector3 import read_from_bytes as build_vector3, Vector3
from battlebit_community_server_api.model.Map import Map
from battlebit_community_server_api.model.Gamemode import Gamemode
from battlebit_community_server_api.model.IpAddress import IpAddress
from battlebit_community_server_api.model.Report import Report
from battlebit_community_server_api.model.ReportReason import ReportReason
from battlebit_community_server_api.model.PlayerBody import PlayerBody
from battlebit_community_server_api.model.ReasonOfDamage import ReasonOfDamage
from battlebit_community_server_api.model.KillReport import KillReport
from battlebit_community_server_api.model.RoundState import RoundState
from battlebit_community_server_api.model.LeaningSide import LeaningSide
from battlebit_community_server_api.model.LoadoutIndex import LoadoutIndex
from battlebit_community_server_api.model.PlayerState import PlayerState
from battlebit_community_server_api.model.MapDayNight import MapDayNight


class TcpParsingService:
    """ The TcpParsingService parses TCP packages in their raw, bytes format and returns data objects. """
    ROUND_SETTINGS_SIZE: int = 33
    MAX_SQUADS_PER_TEAM: int = 64

    @staticmethod
    def split_data_into_individual_messages(data: bytes) -> tuple[tuple[OpCodes, bytes]]:
        """ Takes a bytestring of any length, splits it in individual messages and returns them with OpCodes """
        messages = []
        while len(data) > 0:
            packet_size, data = read_uint32(data)
            message = data[:packet_size]
            data = data[packet_size:]
            op_code_raw, message = read_uint8(message)
            messages.append((OpCodes(op_code_raw), message))
        return tuple(messages)

    @staticmethod
    def parse_player_connected(message: bytes) -> tuple[SteamId, str, IpAddress, Team, Squads, Role]:
        """
        Parses the payload of PLAYER_CONNECTED and

        :return: SteamID, Username, IPAddress, Team, Squad, Role
        """
        steam_id, message = read_uint64(message)
        steam_id = SteamId(steam_id)
        user_name_length, message = read_uint16(message)
        user_name, message = read_string(message, user_name_length)
        ip, message = read_uint32(message)
        ip = IpAddress(ip)
        team, message = read_uint8(message)
        team = Team(team)
        squad, message = read_uint8(message)
        squad = Squads(squad)
        role, message = read_uint8(message)
        role = Role(role)
        return steam_id, user_name, ip, team, squad, role

    @staticmethod
    def parse_player_disconnected(message: bytes) -> SteamId:
        """ Returns SteamID of disconnected player """
        return SteamId(read_uint64(message)[0])

    @staticmethod
    def parse_on_player_typed_message(message: bytes) -> tuple[int, SteamId, ChatChannel, str]:
        """ Returns tuple with messageID, SteamID, ChatChannel and the message string """
        message_id, message = read_uint16(message)
        steam_id, message = read_uint64(message)
        channel, message = read_uint8(message)
        message_length, message = read_uint16(message)
        return message_id, SteamId(steam_id), ChatChannel(channel), message[:message_length].decode("utf-8")

    @staticmethod
    def parse_on_player_joining(message: bytes) -> tuple[SteamId, PlayerStats]:
        """ Returns a tuple with SteamID and PlayerStats """
        steam_id, message = read_uint64(message)
        steam_id = SteamId(steam_id)
        is_banned, message = read_uint8(message)
        is_banned = bool(is_banned)
        roles, message = read_uint64(message)

        progress, message = PlayerProgress.build_player_progress_from_bytes(message)

        tool_progress_size, message = read_uint16(message)
        tool_progress_raw = message[:tool_progress_size]
        message = message[tool_progress_size:]

        achievements_size, message = read_uint16(message)
        achievements_raw = message[:achievements_size]
        message = message[achievements_size:]

        selections_size, message = read_uint16(message)
        selections_raw = message[:selections_size]
        message = message[selections_size:]

        return steam_id, PlayerStats(is_banned=is_banned,
                                     roles=roles,
                                     progress=progress,
                                     tool_progress=tool_progress_raw,
                                     achievements=achievements_raw,
                                     selections=selections_raw)

    @classmethod
    def parse_hail_message(cls, message: bytes) -> ServerInfo:
        """ Parses HAIL """
        # Token
        token_size, message = read_uint16(message)
        token, message = read_string(message, token_size)

        # Version
        version_size, message = read_uint16(message)
        version, message = read_string(message, version_size)

        # Game Port
        game_port, message = read_uint16(message)

        # is port protected
        is_port_protected = struct.unpack("?", message[:1])[0]
        message = message[1:]

        # length of server name
        server_name_length, message = read_uint16(message)

        # Server Name
        server_name = message[:server_name_length].decode("utf-8")
        message = message[server_name_length:]

        # length of game mode name
        game_mode_name_length, message = read_uint16(message)

        # Game Mode Name
        game_mode_name = message[:game_mode_name_length].decode("utf-8")
        message = message[game_mode_name_length:]

        # length of map name
        map_name_length, message = read_uint16(message)

        # Map Name
        map_name = message[:map_name_length].decode("utf-8")
        message = message[map_name_length:]

        # Map Size
        map_size, message = read_uint8(message)

        # DayNight Cycle
        day_night_cycle, message = read_uint8(message)

        # Current Players
        current_players, message = read_uint8(message)

        # Queued Players
        queued_players, message = read_uint8(message)

        # Max Players
        max_players, message = read_uint8(message)

        # Loading screen text length
        loading_screen_text_length, message = read_uint16(message)

        # Loading screen text
        loading_screen_text = message[:loading_screen_text_length].decode("utf-8")
        message = message[loading_screen_text_length:]

        # Server Rules text size
        server_rules_text_length, message = read_uint16(message)

        # Server rules text
        server_rules_text = message[:server_rules_text_length].decode("utf-8")
        message = message[server_rules_text_length:]

        # Round Index
        round_index, message = read_uint32(message)

        # Session ID
        session_id, message = read_uint64(message)

        # Room Settings size
        room_settings_length, message = read_uint32(message)
        # Room Settings
        room_settings = room_settings_from_bytes(message)
        message = message[room_settings_length:]

        # Map & Game mode Rotation size
        map_and_game_mode_rotation_size, message = read_uint32(message)

        # Map & Game mode Rotation
        map_rotation, gamemode_rotation = cls.parse_map_game_mode_rotation(message[:map_and_game_mode_rotation_size])
        message = message[map_and_game_mode_rotation_size:]

        # Round State
        round_state = cls._parse_round_settings(message[:cls.ROUND_SETTINGS_SIZE])
        message = message[cls.ROUND_SETTINGS_SIZE:]

        # Clients
        client_count, message = read_uint8(message)
        clients = []
        for i in range(client_count):
            steam_id, user_name, ip_hash, team, squad, role, is_alive, loadout, wearings, message = \
                cls.parse_hail_message_client_info(message)
            clients.append((steam_id, user_name, ip_hash, team, squad, role, is_alive, loadout, wearings))

        logging.debug("Connected clients:")
        for client in clients:
            logging.debug(f"[{client[0]}] {client[1]}")

        # Squads
        squad_data_size, message = read_uint32(message)
        squad_points_team_us = []
        for _ in range(cls.MAX_SQUADS_PER_TEAM):
            points, message = read_uint32(message)
            squad_points_team_us.append(points)

        squad_points_team_ru = []
        for _ in range(cls.MAX_SQUADS_PER_TEAM):
            points, message = read_uint32(message)
            squad_points_team_ru.append(points)

        if message:
            logging.warning(f"Unparsed HAIL message remains: {message}. "
                            f"Probably an update that broke everything. Please fix.")

        return ServerInfo(name=server_name,
                          version=version,
                          port=game_port,
                          protected=is_port_protected,
                          game_mode=Gamemode(game_mode_name),
                          map_name=Map(map_name),
                          map_size=MapSize(map_size),
                          is_day_mode=day_night_cycle == 0,
                          current_players=current_players,
                          queued_players=queued_players,
                          max_players=max_players,
                          loading_screen_text=loading_screen_text,
                          rules_text=server_rules_text,
                          round_state=round_state,
                          room_settings=room_settings,
                          gamemode_rotation=gamemode_rotation,
                          map_rotation=map_rotation)

    @classmethod
    def parse_on_player_requesting_to_spawn(cls, message: bytes) -> tuple[SteamId, PlayerSpawnArguments, int]:
        """ Returns STEAM-ID, PlayerSpawnArguments and a Vehicle-ID """
        steam_id, message = read_uint64(message)
        steam_id = SteamId(steam_id)
        player_spawning_position, message = read_uint8(message)
        player_spawning_position = PlayerSpawningPosition(player_spawning_position)

        # LOADOUT
        player_loadout, message = build_player_loadout(message)

        # WEARINGS
        player_wearings, message = build_player_wearings(message)
        spawn_position, message = build_vector3(message)
        look_direction, message = build_vector3(message)
        stand, message = read_uint8(message)
        stand = PlayerStand(stand)
        spawn_protection, message = read_float32(message)

        # VEHICLE ID
        vehicle_id, message = read_uint16(message)

        return steam_id, PlayerSpawnArguments(
            position=player_spawning_position,
            loadout=player_loadout,
            wearings=player_wearings,
            vector_position=spawn_position,
            look_direction=look_direction,
            stand=stand,
            spawn_protection=spawn_protection
        ), vehicle_id

    @staticmethod
    def parse_on_player_asking_to_change_role(data: bytes) -> tuple[SteamId, GameRole]:
        steam_id, data = read_uint64(data)
        steam_id = SteamId(steam_id)
        requested_role, data = read_uint8(data)
        requested_role = GameRole(requested_role)
        return steam_id, requested_role

    @staticmethod
    def parse_on_player_asking_to_change_team(data: bytes) -> tuple[SteamId, Team]:
        steam_id, data = read_uint64(data)
        steam_id = SteamId(steam_id)
        requested_team, data = read_uint8(data)
        requested_team = Team(requested_team)
        return steam_id, requested_team

    @staticmethod
    def parse_map_game_mode_rotation(data: bytes) -> tuple[list[Map], list[Gamemode]]:
        maps_in_rotation, data = read_uint32(data)
        maps = []
        for _ in range(maps_in_rotation):
            map_name_size, data = read_uint16(data)
            map_name, data = read_string(data, map_name_size)
            try:
                maps.append(Map(map_name))
            except ValueError:
                logging.warning(f"Unknown map name '{map_name}' found, skipping.")

        game_modes_in_rotation, data = read_uint32(data)
        game_modes = []
        for _ in range(game_modes_in_rotation):
            gamemode_name_size, data = read_uint16(data)
            gamemode_name, data = read_string(data, gamemode_name_size)
            try:
                game_modes.append(Gamemode(gamemode_name))
            except ValueError:
                logging.warning(f"Unknown gamemode name '{gamemode_name}' found, skipping.")

        return maps, game_modes

    @staticmethod
    def parse_notify_new_map_rotation(data: bytes) -> list[Map]:
        map_count, data = read_uint32(data)
        maps = []
        for _ in range(map_count):
            map_name_size, data = read_uint16(data)
            map_name, data = read_string(data, map_name_size)
            maps.append(Map(map_name))
        return maps

    @staticmethod
    def parse_notify_new_gamemode_rotation(data: bytes) -> list[Gamemode]:
        """ ToDo: Currently broken on server-side, implement when fixed. """
        raise NotImplementedError

    @staticmethod
    def parse_on_player_changed_role(data: bytes) -> tuple[SteamId, GameRole]:
        """ Returns SteamID and the players new game role """
        steam_id, data = read_uint64(data)
        steam_id = SteamId(steam_id)
        new_role, data = read_uint8(data)
        new_role = GameRole(new_role)
        return steam_id, new_role

    @staticmethod
    def parse_on_player_joined_a_squad(data: bytes) -> tuple[SteamId, Squads, bool]:
        """
        Parses the payload from ON_PLAYER_JOINED_A_SQUAD

        :return: SteamID, new Squad of the player and a bool that tells if the player is the new captain of that squad
        """
        steam_id, data = read_uint64(data)
        steam_id = SteamId(steam_id)
        new_squad, data = read_uint8(data)
        new_squad = Squads(new_squad)
        is_squad_captain, _ = read_bool(data)
        return steam_id, new_squad, is_squad_captain

    @staticmethod
    def parse_on_player_left_squad(data: bytes) -> SteamId:
        return SteamId(read_uint64(data)[0])

    @staticmethod
    def parse_on_player_changed_team(data: bytes) -> tuple[SteamId, Team]:
        steam_id, new_team = read_uint64(data)
        steam_id, new_team = SteamId(steam_id), Team(read_uint8(new_team)[0])
        return steam_id, new_team

    @staticmethod
    def parse_on_player_report(data: bytes) -> Report:
        # Untested, should work though
        reporter, data = read_uint64(data)
        reported, data = read_uint64(data)
        reason, data = read_uint8(data)
        if data:
            additional_info_size, data = read_uint16(data)
            additional_info, _ = read_string(data, additional_info_size)
        else:
            additional_info = None
        return Report(SteamId(reporter), SteamId(reported), ReportReason(reason), additional_info)

    @staticmethod
    def parse_on_player_killed_another_player(data: bytes) -> KillReport:
        killer, data = read_uint64(data)
        killer = SteamId(killer)
        killer_pos, data = build_vector3(data)
        victim, data = read_uint64(data)
        victim = SteamId(victim)
        victim_pos, data = build_vector3(data)
        tool_size, data = read_uint16(data)
        tool_name, data = read_string(data, tool_size)
        if not tool_name:
            tool_name = None
        body_part, data = read_uint8(data)
        body_part = PlayerBody(body_part)
        damage_reason, _ = read_uint8(data)
        damage_reason = ReasonOfDamage(damage_reason)

        return KillReport(killer, killer_pos, victim, victim_pos, tool_name, body_part, damage_reason)

    @classmethod
    def parse_save_player_stats(cls, data: bytes) -> tuple[SteamId, PlayerStats]:
        """ Returns SteamID and Stats of player """
        # Data is identical, so we can just use the parser for ON_PLAYERS_JOINING
        return cls.parse_on_player_joining(data)

    @classmethod
    def parse_on_player_given_up(cls, data: bytes) -> SteamId:
        """ Returns SteamID of player that gave up"""
        return SteamId(read_uint64(data)[0])

    @staticmethod
    def parse_on_player_spawn(data: bytes) -> tuple[SteamId, PlayerLoadout, PlayerWearings, Vector3]:
        """ Returns SteamID, Loadout, Wearings and spawn position of spawning player """
        steam_id, data = read_uint64(data)
        steam_id = SteamId(steam_id)
        player_loadout, data = build_player_loadout(data)
        player_wearings, data = build_player_wearings(data)
        spawn_pos, _ = build_vector3(data)
        return steam_id, player_loadout, player_wearings, spawn_pos

    @staticmethod
    def parse_on_player_die(data: bytes) -> SteamId:
        """ Returns the SteamID of the player that died """
        return SteamId(read_uint64(data)[0])

    @staticmethod
    def parse_on_player_revived_another(data: bytes) -> tuple[SteamId, SteamId]:
        """ Returns the SteamID of the reviver and the revived. First entry is the reviver. """
        reviving_player, data = read_uint64(data)
        reviving_player = SteamId(reviving_player)
        revived_player, _ = read_uint64(data)
        revived_player = SteamId(revived_player)
        return reviving_player, revived_player

    @classmethod
    def parse_new_round_state(cls, data: bytes) -> RoundState:
        return cls._parse_round_settings(data)

    @staticmethod
    def _parse_round_settings(round_settings_bytes: bytes) -> RoundState:
        state, round_settings_bytes = read_uint8(round_settings_bytes)
        state = GameState(state)
        team_a_tickets, round_settings_bytes = read_uint64(round_settings_bytes)
        team_b_tickets, round_settings_bytes = read_uint64(round_settings_bytes)
        max_tickets, round_settings_bytes = read_uint64(round_settings_bytes)
        players_to_start, round_settings_bytes = read_uint32(round_settings_bytes)
        seconds_left, round_settings_bytes = read_uint32(round_settings_bytes)
        return RoundState(state, team_a_tickets, team_b_tickets, max_tickets, players_to_start, seconds_left)

    @staticmethod
    def parse_game_tick(data: bytes) -> tuple[int, list[PlayerState]]:
        """ Returns amount of players ALIVE as well as the state of everyone alive """
        decompress_x, data = read_float32(data)
        decompress_y, data = read_float32(data)
        decompress_z, data = read_float32(data)

        offset_x, data = read_float32(data)
        offset_y, data = read_float32(data)
        offset_z, data = read_float32(data)

        player_count, data = read_uint8(data)
        player_states = []
        for player in range(player_count):
            steam_id, data = read_uint64(data)
            steam_id = SteamId(steam_id)
            com_pos_x, data = read_uint16(data)
            com_pos_y, data = read_uint16(data)
            com_pos_z, data = read_uint16(data)
            player_position = Vector3(
                (com_pos_x * decompress_x) - offset_x,
                (com_pos_y * decompress_y) - offset_y,
                (com_pos_z * decompress_z) - offset_z,
            )
            # FixMe: Unclear how player_health value is to be interpreted, leaving it out for now
            player_health, data = read_uint8(data)
            player_health = (player_health * 0.5) - 1.0
            player_standing, data = read_uint8(data)
            player_standing = PlayerStand(player_standing)
            leaning_side, data = read_uint8(data)
            leaning_side = LeaningSide(leaning_side)
            loadout_index, data = read_uint8(data)
            loadout_index = LoadoutIndex(loadout_index)
            in_seat, data = read_bool(data)
            is_bleeding, data = read_bool(data)
            ping, data = read_uint16(data)
            player_states.append(
                PlayerState(
                    steam_id, player_position, player_standing, leaning_side, loadout_index, in_seat, is_bleeding, ping
                )
            )

        return player_count, player_states

    @staticmethod
    def parse_on_squad_points_changed(data: bytes) -> tuple[Team, Squads, int]:
        """ Returns the squads current points """
        team, data = read_uint8(data)
        team = Team(team)
        squad, data = read_uint8(data)
        squad = Squads(squad)
        points, _ = read_uint32(data)
        return team, squad, points

    @staticmethod
    def parse_notify_new_round_id(data: bytes) -> tuple[int, int]:
        """ Returns Round-Index and SessionID """
        round_index, data = read_uint32(data)
        session_id, _ = read_uint64(data)
        return round_index, session_id

    @staticmethod
    def parse_log(data: bytes) -> str:
        """ Returns log message """
        log_msg_size, data = read_uint16(data)
        return read_string(data, log_msg_size)[0]

    @staticmethod
    def parse_on_squad_leader_changed(data: bytes) -> tuple[SteamId, Squads]:
        """ Returns SteamID of the player who is the new squad leader, and the squad (without team info) """
        steam_id, data = read_uint64(data)
        steam_id = SteamId(steam_id)
        squad, _ = read_uint8(data)
        squad = Squads(squad)
        return steam_id, squad

    @staticmethod
    def parse_update_connected_players(data: bytes) -> tuple[int, int, int]:
        """ Returns current players, players in queue and max players, in that order. """
        current_players, data = read_uint8(data)
        in_queue, data = read_uint8(data)
        max_players, data = read_uint8(data)  # Intentionally whacky, ask Oki
        max_players, data = read_uint8(data)
        return current_players, in_queue, max_players

    @staticmethod
    def parse_update_new_game_data(data: bytes) -> tuple[int, int, int, str, MapSize, str, MapDayNight]:
        """
        Returns (in that order):
        Current players, players in queue, max players, game mode name, map size, map name, day-night flag
        """
        # ToDo: Maybe make this into a dataclass later.
        current_players, data = read_uint8(data)
        in_queue, data = read_uint8(data)
        max_players, data = read_uint8(data)  # Intentionally whacky, ask Oki
        max_players, data = read_uint8(data)
        gamemode_name_len, data = read_uint16(data)
        gamemode_name, data = read_string(data, gamemode_name_len)
        map_size, data = read_uint8(data)
        map_size = MapSize(map_size)
        map_name_len, data = read_uint16(data)
        map_name, data = read_string(data, map_name_len)
        game_type, data = read_uint8(data)
        day_night, _ = read_uint8(data)
        day_night = MapDayNight(day_night)
        return current_players, in_queue, max_players, gamemode_name, map_size, map_name, day_night

    @staticmethod
    def parse_hail_message_client_info(client_info: bytes) -> tuple[SteamId, str, int, Team, Squads, Role, bool,
                                                                    Optional[PlayerLoadout], Optional[PlayerWearings],
                                                                    bytes]:
        """
        Parses the guaranteed 24 bytes segments of client info, returned by the HAIL message.
        Then, if the player is alive, parses the PlayerLoadout and PlayerWearing.

        :return: Tuple containing: Steam-ID, Username, IP Hash, Team, Squad, Role, IsAliveStatus, PlayerLoadout,
                                   PlayerWearing and the remainder of buffer.
        :note: If the player is not alive, PlayerLoadout and PlayerWearing are None.
        """
        steam_id, client_info = read_uint64(client_info)
        steam_id = SteamId(steam_id)
        user_name_size, client_info = read_uint16(client_info)
        user_name, client_info = read_string(client_info, user_name_size)
        ip_hash, client_info = read_uint32(client_info)
        team, client_info = read_uint8(client_info)
        team = Team(team)
        squad, client_info = read_uint8(client_info)
        squad = Squads(squad)
        role, client_info = read_uint8(client_info)
        role = Role(role)
        is_alive, client_info = read_uint8(client_info)
        is_alive = bool(is_alive)
        if is_alive:
            _, client_info = read_uint32(client_info)  # Loadout size
            player_loadout, client_info = build_player_loadout(client_info)
            player_wearings, client_info = build_player_wearings(client_info)
        else:
            player_loadout = None
            player_wearings = None

        return steam_id, user_name, ip_hash, team, squad, role, is_alive, player_loadout, player_wearings, client_info



from dataclasses import dataclass
from typing import Optional

from battlebit_community_server_api.model.SteamId import SteamId
from battlebit_community_server_api.model.ReportReason import ReportReason


@dataclass
class Report:
    reporter: SteamId
    reported: SteamId
    reason: ReportReason
    additional_info: Optional[str] = None



from enum import StrEnum


class Gamemode(StrEnum):
    CONQUEST = "CONQ"
    INFANTRY_CONQUEST = "INFCONQ"
    DOMINATION = "DOMI"
    RUSH = "RUSH"
    CAPTURE_THE_FLAG = "CTF"
    FRONTLINE = "FRONTLINE"
    TEAM_DEATHMATCH = "TDM"
    ELIMINATION = "ELI"



import logging
import struct


def read_uint8(buf: bytes) -> tuple[int, bytes]:
    """ Reads uint8, returns remaining buffer """
    uint8 = int(struct.unpack("B", buf[:1])[0])
    return uint8, buf[1:]


def write_uint8(d: int) -> bytes:
    """ Writes an integer to a single byte """
    if not 0 <= d <= 255:
        raise ValueError(f"Can not convert integer {d} to uint8!")
    return struct.pack("B", d)


def read_bool(buf: bytes) -> tuple[bool, bytes]:
    """ Reads uint8, converts result to boolean, returns remaining buffer """
    uint8 = int(struct.unpack("B", buf[:1])[0])
    return bool(uint8), buf[1:]


def write_bool(d: bool) -> bytes:
    """ Writes a boolean to a single byte """
    return struct.pack("B", d)


def read_uint16(buf: bytes) -> tuple[int, bytes]:
    """ Reads uint16, returns remaining buffer """
    uint16 = int(struct.unpack("H", buf[:2])[0])
    return uint16, buf[2:]


def read_uint32(buf: bytes) -> tuple[int, bytes]:
    """ Reads uint32, returns remaining buffer """
    uint32 = int(struct.unpack("I", buf[:4])[0])
    return uint32, buf[4:]


def read_uint64(buf: bytes) -> tuple[int, bytes]:
    """ Reads uint64, returns remaining buffer """
    uint64 = int(struct.unpack("Q", buf[:8])[0])
    return uint64, buf[8:]


def read_float32(buf: bytes) -> tuple[float, bytes]:
    f_32 = struct.unpack("f", buf[:4])[0]
    return f_32, buf[4:]


def write_float32(d: float) -> bytes:
    return struct.pack("f", d)


def read_string(buf: bytes, length: int) -> tuple[str, bytes]:
    """ Reads string of specified length (from 0), returns remaining buffer """
    return buf[:length].decode("utf-8"), buf[length:]


def write_string(string_to_write: str) -> bytes:
    """ writes string to bytes, including the 2-Byte length prefix """
    b = struct.pack("H", len(string_to_write))
    b += string_to_write.encode("utf-8")
    return b



from enum import IntEnum


class PlayerSpawningPosition(IntEnum):
    SPAWN_AT_POINT = 0
    SPAWN_AT_RALLY = 1
    SPAWN_AT_FRIEND = 2
    SPAWN_AT_VEHICLE = 3
    NONE = 4

    def as_bytes(self) -> bytes:
        return self.to_bytes(length=1, byteorder="little")



from enum import StrEnum


class Map(StrEnum):
    AZAGOR = "Azagor"
    BASRA = "Basra"
    CONSTRUCTION = "Construction"
    DISTRICT = "District"
    DUSTY_DEW = "Dustydew"
    EDUARDOVO = "Eduardovo"
    FRUGIS = "Frugis"
    ISLE = "Isle"
    LONOVO = "Lonovo"
    MULTU_ISLANDS = "MultuIslands"
    NAMAK = "Namak"
    OIL_DUNES = "OilDunes"
    RIVER = "River"
    SALHAN = "Salhan"
    SANDY_SUNSET = "SandySunset"
    TENSA_TOWN = "TensaTown"
    VALLEY = "Valley"
    WAKISTAN = "Wakistan"
    WINE_PARADISE = "WineParadise"
    OLD_MULTU_ISLANDS = "Old_MultuIslands"
    OLD_DISTRICT = "Old_District"
    OLD_EDUARDOVO = "Old_Eduardovo"
    OLD_NAMAK = "Old_Namak"
    OLD_OIL_DUNES = "Old_OilDunes"



from dataclasses import dataclass

from battlebit_community_server_api.helper.StructHelper import *


@dataclass
class PlayerWearings:
    head: str
    chest: str
    belt: str
    backpack: str
    eye: str
    face: str
    hair: str
    skin: str
    uniform: str
    camo: str

    def as_bytes(self) -> bytes:
        out = write_string(self.head)
        out += write_string(self.chest)
        out += write_string(self.belt)
        out += write_string(self.backpack)
        out += write_string(self.eye)
        out += write_string(self.face)
        out += write_string(self.hair)
        out += write_string(self.skin)
        out += write_string(self.uniform)
        out += write_string(self.camo)
        return out


def read_from_bytes(data: bytes) -> tuple[PlayerWearings, bytes]:
    head_size, data = read_uint16(data)
    head, data = read_string(data, head_size)

    chest_size, data = read_uint16(data)
    chest, data = read_string(data, chest_size)

    belt_size, data = read_uint16(data)
    belt, data = read_string(data, belt_size)

    backpack_size, data = read_uint16(data)
    backpack, data = read_string(data, backpack_size)

    eye_size, data = read_uint16(data)
    eye, data = read_string(data, eye_size)

    face_size, data = read_uint16(data)
    face, data = read_string(data, face_size)

    hair_size, data = read_uint16(data)
    hair, data = read_string(data, hair_size)

    skin_size, data = read_uint16(data)
    skin, data = read_string(data, skin_size)

    uniform_size, data = read_uint16(data)
    uniform, data = read_string(data, uniform_size)

    camo_size, data = read_uint16(data)
    camo, data = read_string(data, camo_size)

    return PlayerWearings(head=head,
                          chest=chest,
                          belt=belt,
                          backpack=backpack,
                          eye=eye,
                          face=face,
                          hair=hair,
                          skin=skin,
                          uniform=uniform,
                          camo=camo), data



import asyncio
import logging
import time
from threading import Thread

from battlebit_community_server_api.ApiServer import ApiServer, build_api_server
from battlebit_community_server_api.command.Command import SetMapRotation
from battlebit_community_server_api.model.Map import Map

event_loop = asyncio.new_event_loop()
my_api_server: ApiServer = build_api_server("0.0.0.0", 30000, event_loop)

# Construct new map rotation
new_map_rotation = [
    Map.DUSTY_DEW,
    Map.MULTU_ISLANDS,
    Map.OIL_DUNES,
    Map.SANDY_SUNSET
]


# Prepare command
my_command = SetMapRotation(new_map_rotation)


logging.basicConfig(level=logging.DEBUG)  # Optional: If you want to have an in-depth view into what's going on.

# Adding Command to the queue from outside the main loop (triggered by surrounding code)
# required the API to be run in a Thread, so it doesn't block.
api_thread = Thread(target=my_api_server.start, daemon=False)
api_thread.start()
SLEEP_TIME = 3
for i in range(SLEEP_TIME):
    logging.debug(f"Sleeping: {SLEEP_TIME - i}")
    time.sleep(1)

logging.debug(f"Adding SetMapRotation command to queue")

# It is PARAMOUNT that all methods on the API server are called threadsafe. Otherwise, you might
# run into race conditions or locks. Use asyncio.run_coroutine_threadsafe whenever possible !
asyncio.run_coroutine_threadsafe(
    my_api_server.add_command_to_queue(my_command),
    event_loop
)



import asyncio
import logging
from asyncio import AbstractEventLoop
from typing import Callable, Optional

from battlebit_community_server_api.command.Command import Command
from battlebit_community_server_api.model.OpCodes import OpCodes
from battlebit_community_server_api.service.GameServerCommunicationService import GameServerCommunicationService
from battlebit_community_server_api.model.OutgoingGameServerMessage import OutgoingGameServerMessage


class ApiServer:
    def __init__(self, game_server_communication_service: GameServerCommunicationService,
                 event_loop: Optional[AbstractEventLoop]) -> None:
        self._game_server_communication_service = game_server_communication_service
        # ToDo: This is new in Python 3.10, add a switch for it, if that's the only breaking change to 3.7
        self._main_loop = event_loop

    def start(self) -> None:
        if not self._main_loop:
            self._main_loop = asyncio.new_event_loop()
        self._main_loop.run_until_complete(self._game_server_communication_service.start())

    def register_handler(self, op_code: OpCodes, handler: Callable) -> None:
        self._game_server_communication_service.register_handler(op_code, handler)

    async def add_command_to_queue(self, command: Command) -> None:
        await self._game_server_communication_service.add_command_to_queue(command)

    async def add_outgoing_game_server_message_to_queue(self, msg: OutgoingGameServerMessage) -> None:
        await self._game_server_communication_service.add_outgoing_game_server_message_to_queue(msg)

    @property
    def loop(self) -> AbstractEventLoop:
        return self._main_loop

    @loop.setter
    def loop(self, new_loop: AbstractEventLoop) -> None:
        self._main_loop = new_loop


def build_api_server(host: str, port: int, event_loop: Optional[AbstractEventLoop]) -> ApiServer:
    return ApiServer(
        GameServerCommunicationService(
            host, port
        ), event_loop
    )


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    api_server = build_api_server("0.0.0.0", 30000, None)
    logging.debug("Starting BattleBit Community Server Python API")
    api_server.start()



from dataclasses import dataclass

from battlebit_community_server_api.helper.StructHelper import *


@dataclass
class RoomSettings:
    damage_multiplier: float
    friendly_fire_enabled: bool
    hide_map_votes: bool
    only_winner_team_can_vote: bool
    player_collision_enabled: bool

    medic_limit_per_squad: int
    engineer_limit_per_squad: int
    support_limit_per_squad: int
    recon_limit_per_squad: int

    can_vote_for_day: bool
    can_vote_for_night: bool

    tank_spawn_delay_multiplier: float
    transport_spawn_delay_multiplier: float
    sea_vehicle_spawn_delay_multiplier: float
    apc_spawn_delay_multiplier: float
    helicopter_spawn_delay_multiplier: float

    unlock_all_attachments: bool
    teamless_mode_enabled: bool
    squad_required_to_change_role: bool

    def to_bytes(self) -> bytes:
        out = write_float32(self.damage_multiplier)
        out += write_bool(self.friendly_fire_enabled)
        out += write_bool(self.hide_map_votes)
        out += write_bool(self.only_winner_team_can_vote)
        out += write_bool(self.player_collision_enabled)

        out += write_uint8(self.medic_limit_per_squad)
        out += write_uint8(self.engineer_limit_per_squad)
        out += write_uint8(self.support_limit_per_squad)
        out += write_uint8(self.recon_limit_per_squad)

        out += write_bool(self.can_vote_for_day)
        out += write_bool(self.can_vote_for_night)

        out += write_float32(self.tank_spawn_delay_multiplier)
        out += write_float32(self.transport_spawn_delay_multiplier)
        out += write_float32(self.sea_vehicle_spawn_delay_multiplier)
        out += write_float32(self.apc_spawn_delay_multiplier)
        out += write_float32(self.helicopter_spawn_delay_multiplier)

        out += write_bool(self.unlock_all_attachments)
        out += write_bool(self.teamless_mode_enabled)
        out += write_bool(self.squad_required_to_change_role)
        return out


def from_bytes(data: bytes) -> RoomSettings:
    damage_multiplier, data = read_float32(data)
    friendly_fire_enabled, data = read_bool(data)
    hide_map_votes, data = read_bool(data)
    only_winner_team_can_vote, data = read_bool(data)
    player_collision_enabled, data = read_bool(data)

    medic_limit_per_squad, data = read_uint8(data)
    engineer_limit_per_squad, data = read_uint8(data)
    support_limit_per_squad, data = read_uint8(data)
    recon_limit_per_squad, data = read_uint8(data)

    can_vote_for_day, data = read_bool(data)
    can_vote_for_night, data = read_bool(data)

    tank_spawn_delay_multiplier, data = read_float32(data)
    transport_spawn_delay_multiplier, data = read_float32(data)
    sea_vehicle_spawn_delay_multiplier, data = read_float32(data)
    apc_spawn_delay_multiplier, data = read_float32(data)
    helicopter_spawn_delay_multiplier, data = read_float32(data)

    unlock_all_attachments, data = read_bool(data)
    teamless_mode_enabled, data = read_bool(data)
    squad_required_to_change_role, data = read_bool(data)

    return RoomSettings(damage_multiplier, friendly_fire_enabled, hide_map_votes, only_winner_team_can_vote,
                        player_collision_enabled, medic_limit_per_squad, engineer_limit_per_squad,
                        support_limit_per_squad, recon_limit_per_squad, can_vote_for_day, can_vote_for_night,
                        tank_spawn_delay_multiplier, transport_spawn_delay_multiplier,
                        sea_vehicle_spawn_delay_multiplier, apc_spawn_delay_multiplier,
                        helicopter_spawn_delay_multiplier, unlock_all_attachments, teamless_mode_enabled,
                        squad_required_to_change_role)



from enum import IntEnum


class ReportReason(IntEnum):
    CHEATING_WALLHACK = 0
    CHEATING_AIMBOT = 1
    CHEATING_SPEEDHACK = 2
    RACISM_DISCRIMINATION_TEXT = 3
    RACISM_DISCRIMINATION_VOICE = 4
    SPAMMING_TEXT = 5
    SPAMMING_VOICE = 6
    BAD_LANGUAGE_TEXT = 7
    BAD_LANGUAGE_VOICE = 8
    GRIEFING = 9
    EXPLOITING = 10
    OTHER_TOXIC_BEHAVIOUR = 11
    USER_PROFILE_NAME_PICTURE = 12



import logging
from datetime import datetime

from battlebit_community_server_api.ApiServer import ApiServer, build_api_server
from battlebit_community_server_api.command.Command import SayToAllChat, ForceStartGame, SayToChat
from battlebit_community_server_api.handler.DefaultHandlers import DefaultHandlers
from battlebit_community_server_api.model.OpCodes import OpCodes
from battlebit_community_server_api.service.TcpParsingService import TcpParsingService

my_api_server: ApiServer = build_api_server("0.0.0.0", 30000, None)


async def chat_messages_handler(data: bytes) -> bytes:
    """
    This function:

    1. Parses the data received by the "On Player Typed Message" event.
    2. Builds a new message depending on what the content of the original was.
    3. Sends the new message to chat.
    4. Executes the default handler for "On Player Typed Message" event.
    """
    message_id, steam_id, channel, message = TcpParsingService.parse_on_player_typed_message(data)
    if message.startswith("!"):
        message = message.lstrip("!").lower()
        if message == "time":
            response_to_chat = SayToAllChat(f"Current time: {datetime.now().strftime('%H:%M:%S')}")
        elif message == "steamid":
            response_to_chat = SayToChat(f"Your STEAM-ID: {steam_id}", steam_id)
        elif message == "forcestart":
            response_to_chat = SayToAllChat(f"Force starting round")
            await my_api_server.add_command_to_queue(ForceStartGame())
        else:
            response_to_chat = SayToAllChat(f"Sorry, I couldn't recognize your command :(")

        await my_api_server.add_command_to_queue(response_to_chat)

    return await DefaultHandlers.get_default_handler_by_op_code(OpCodes.ON_PLAYER_TYPED_MESSAGE)(data)


logging.basicConfig(level=logging.DEBUG)  # Optional: If you want to have an in-depth view into what's going on.
my_api_server.register_handler(OpCodes.ON_PLAYER_TYPED_MESSAGE, chat_messages_handler)

my_api_server.start()



import logging
import struct
from typing import Callable

from battlebit_community_server_api.command.Command import SetRoleTo, SetTeamTo
from battlebit_community_server_api.model.OpCodes import OpCodes
from battlebit_community_server_api.model.OutgoingGameServerMessage import OutgoingGameServerMessage
from battlebit_community_server_api.model.PlayerJoiningArguments import PlayerJoiningArguments
from battlebit_community_server_api.model.Squads import Squads
from battlebit_community_server_api.model.Team import Team
from battlebit_community_server_api.service.TcpParsingService import TcpParsingService


class DefaultHandlers:
    @classmethod
    def get_default_handler_by_op_code(cls, op_code: OpCodes) -> Callable:
        if op_code == OpCodes.NONE:
            return cls.no_response
        elif op_code == OpCodes.HAIL:
            return cls.hail_handler
        elif op_code == OpCodes.ACCEPTED:
            return cls.no_response
        elif op_code == OpCodes.DENIED:
            return cls.no_response
        elif op_code == OpCodes.EXECUTE_COMMAND:
            return cls.no_response
        elif op_code == OpCodes.SEND_PLAYER_STATS:
            return cls.no_response
        elif op_code == OpCodes.SPAWN_PLAYER:
            return cls.no_response
        elif op_code == OpCodes.SET_NEW_ROOM_SETTINGS:
            return cls.no_response
        elif op_code == OpCodes.RESPOND_PLAYER_MESSAGE:
            return cls.no_response
        elif op_code == OpCodes.SET_NEW_ROUND_STATE:
            return cls.no_response
        elif op_code == OpCodes.SET_PLAYER_WEAPON:
            return cls.no_response
        elif op_code == OpCodes.SET_PLAYER_GADGET:
            return cls.no_response
        elif op_code == OpCodes.PLAYER_CONNECTED:
            return cls.no_response
        elif op_code == OpCodes.PLAYER_DISCONNECTED:
            return cls.no_response
        elif op_code == OpCodes.ON_PLAYER_TYPED_MESSAGE:
            return cls.on_player_typed_message_handler
        elif op_code == OpCodes.ON_PLAYER_KILLED_ANOTHER_PLAYER:
            return cls.no_response
        elif op_code == OpCodes.ON_PLAYER_JOINING:
            return cls.on_player_joining_handler
        elif op_code == OpCodes.SAVE_PLAYER_STATS:
            return cls.no_response
        elif op_code == OpCodes.ON_PLAYER_ASKING_TO_CHANGE_ROLE:
            return cls.on_player_asking_to_change_role
        elif op_code == OpCodes.ON_PLAYER_CHANGED_ROLE:
            return cls.no_response
        elif op_code == OpCodes.ON_PLAYER_JOINED_A_SQUAD:
            return cls.no_response
        elif op_code == OpCodes.ON_PLAYER_LEFT_SQUAD:
            return cls.no_response
        elif op_code == OpCodes.ON_PLAYER_CHANGED_TEAM:
            return cls.no_response
        elif op_code == OpCodes.ON_PLAYER_REQUESTING_TO_SPAWN:
            return cls.on_player_requesting_to_spawn
        elif op_code == OpCodes.ON_PLAYER_REPORT:
            return cls.no_response
        elif op_code == OpCodes.ON_PLAYER_SPAWN:
            return cls.no_response
        elif op_code == OpCodes.ON_PLAYER_DIE:
            return cls.no_response
        elif op_code == OpCodes.NOTIFY_NEW_MAP_ROTATION:
            return cls.no_response
        elif op_code == OpCodes.NOTIFY_NEW_GAME_MODE_ROTATION:
            return cls.no_response
        elif op_code == OpCodes.NOTIFY_NEW_ROUND_STATE:
            return cls.no_response
        elif op_code == OpCodes.ON_PLAYER_ASKING_TO_CHANGE_TEAM:
            return cls.on_player_asking_to_change_team
        elif op_code == OpCodes.GAME_TICK:
            return cls.no_response
        elif op_code == OpCodes.ON_PLAYER_GIVEN_UP:
            return cls.no_response
        elif op_code == OpCodes.ON_PLAYER_REVIVED_ANOTHER:
            return cls.no_response
        elif op_code == OpCodes.ON_SQUAD_POINTS_CHANGED:
            return cls.no_response
        elif op_code == OpCodes.NOTIFY_NEW_ROUND_ID:
            return cls.no_response
        elif op_code == OpCodes.LOG:
            return cls.log
        elif op_code == OpCodes.ON_SQUAD_LEADER_CHANGED:
            return cls.no_response
        elif op_code == OpCodes.UPDATE_NEW_GAME_DATA:
            return cls.no_response
        elif op_code == OpCodes.UPDATE_CONNECTED_PLAYERS:
            return cls.no_response
        raise NotImplementedError(f"Could not find default handler for OP code {OpCodes(op_code).name}")

    @classmethod
    async def hail_handler(cls, data: bytes) -> bytes:
        logging.debug("Accepting HAIL...")
        return OutgoingGameServerMessage(op_code=OpCodes.ACCEPTED).serialize(without_packet_size=True)

    @classmethod
    async def on_player_joining_handler(cls, data: bytes) -> bytes:
        steam_id, player_stats = TcpParsingService.parse_on_player_joining(data)
        response = OutgoingGameServerMessage(op_code=OpCodes.SEND_PLAYER_STATS,
                                             value=steam_id.to_bytes()
                                             )
        response.add_bytes(PlayerJoiningArguments(player_stats, squad=Squads.NO_SQUAD, team=Team.NONE).to_bytes())
        return response.serialize()

    @classmethod
    async def on_player_typed_message_handler(cls, data: bytes) -> bytes:
        message_id, steam_id, channel, message = TcpParsingService.parse_on_player_typed_message(data)
        message_accepted = True
        response = OutgoingGameServerMessage(op_code=OpCodes.RESPOND_PLAYER_MESSAGE,
                                             value=struct.pack("H", message_id) + struct.pack("B", message_accepted))
        return response.serialize()

    @classmethod
    async def on_player_requesting_to_spawn(cls, data: bytes) -> bytes:
        steam_id, spawn_arguments, vehicle_id = TcpParsingService.parse_on_player_requesting_to_spawn(data)
        response = OutgoingGameServerMessage(op_code=OpCodes.SPAWN_PLAYER)
        response.add_bytes(steam_id.to_bytes())
        response.add_bytes(struct.pack("B", True))
        response.add_bytes(spawn_arguments.to_bytes())
        response.add_bytes(struct.pack("H", vehicle_id))
        return response.serialize()

    @classmethod
    async def on_player_asking_to_change_role(cls, data: bytes) -> bytes:
        steam_id, requested_role = TcpParsingService.parse_on_player_asking_to_change_role(data)
        response_command = SetRoleTo(steam_id, requested_role)
        response = OutgoingGameServerMessage(op_code=OpCodes.EXECUTE_COMMAND)
        response.add_bytes(struct.pack("H", len(response_command)))
        response.add_string(response_command.as_string())
        return response.serialize()

    @classmethod
    async def on_player_asking_to_change_team(cls, data: bytes) -> bytes:
        steam_id, requested_team = TcpParsingService.parse_on_player_asking_to_change_team(data)
        response_command = SetTeamTo(steam_id, requested_team)
        response = OutgoingGameServerMessage(op_code=OpCodes.EXECUTE_COMMAND)
        response.add_bytes(struct.pack("H", len(response_command)))
        response.add_string(response_command.as_string())
        return response.serialize()

    @classmethod
    async def no_response(cls, _: bytes) -> None:
        """ Specialized handler that signals that there is no need to respond """
        return

    @classmethod
    async def log(cls, data: bytes) -> None:
        logging.debug(TcpParsingService.parse_log(data))
        return

    @classmethod
    async def debug_log_bytes(cls, data: bytes) -> None:
        """ Debug handler to log certain OP codes """
        logging.debug(data)
        return



from dataclasses import dataclass

from battlebit_community_server_api.model.GameState import GameState


@dataclass
class RoundState:
    game_state: GameState
    team_a_tickets: int
    team_b_tickets: int
    max_tickets: int
    players_to_start: int
    seconds_left: int



import struct
from dataclasses import dataclass

from battlebit_community_server_api.model import Role
from battlebit_community_server_api.model.PlayerProgress import PlayerProgress


@dataclass
class PlayerStats:
    is_banned: bool
    roles: Role
    progress: PlayerProgress
    tool_progress: bytes
    achievements: bytes
    selections: bytes

    def to_bytes(self) -> bytes:
        out = struct.pack("B", self.is_banned)
        out += struct.pack("Q", self.roles)
        out += self.progress.to_bytes()
        out += struct.pack("H", len(self.tool_progress))
        out += self.tool_progress
        out += struct.pack("H", len(self.achievements))
        out += self.achievements
        out += struct.pack("H", len(self.selections))
        out += self.selections
        return out



from enum import IntEnum


class WeaponType(IntEnum):
    RIFLE = 0
    DMR = 1
    SNIPER_RIFLE = 2
    LIGHT_SUPPORT_GUN = 3
    LIGHT_MACHINE_GUN = 4
    SUB_MACHINE_GUN = 5
    PISTOL = 6
    AUTO_PISTOL = 7
    HEAVY_PISTOL = 8
    CARBINE = 9
    PERSONAL_DEFENSE_WEAPON = 10



import struct
from typing import Optional

from battlebit_community_server_api.model.OpCodes import OpCodes


class OutgoingGameServerMessage:
    """ Abstraction for building outgoing messages to the game server """
    def __init__(self, op_code: OpCodes, value: Optional[bytes] = None) -> None:
        self._op_code = op_code
        self._value: bytes = value if value else bytes()

    def add_bytes(self, new_bytes: bytes) -> None:
        self._value += new_bytes

    def add_string(self, string: str) -> None:
        self._value += string.encode("utf-8")

    def serialize(self, without_packet_size: bool = False) -> bytes:
        serialized_message = struct.pack("B", self._op_code)
        serialized_message += self._value
        if without_packet_size:
            return serialized_message
        return struct.pack("I", len(serialized_message)) + serialized_message
